#! /usr/local/bin/python3
# -*- mode: Python3; tab-width: 4; indent-tabs-mode: nil -*-

"""
This program uploads a data file (.xlsx) from client, creates a temp directory
to store it in; calls another module to create an output data file (or 
update the input file?), passing back its path; then sends that output
data back as the response.

Logging information, including warnings and errors, normally go on a separate sheet of the returned .xlsx file.
But if we get an error that means .xlsx cannot be created, we need to return instead an HTML page explaining the error a little.

"""

import cgi
import html
from shutil import copyfileobj
from sys import stdout
from tempfile import TemporaryDirectory
from traceback import format_exc

from pontowb import WB

class Log:

    def __init__(self):
        self.messages = []

    def info(self, message):
        #print('Info: ', message, file=stderr)
        self.messages.append(('Info', message))

    def warn(self, message):
        #print('Warning: ', message, file=stderr)
        self.messages.append(('Warning', message))

    def error(self, message):
        #print('Error: ', message, file=stderr)
        self.messages.append(('Error', message))

    def debug(self, message):
        pass
        #print('Debug: ', message, file=stderr)

    def exception(self, intro):
        #print('Error: ', intro, file=stderr)
        #print_exc(file=stderr)
        formatted_lines = format_exc().splitlines()
        self.error(formatted_lines[-1])
        self.error(formatted_lines[-2])

    def all_messages(self):
        for message in self.messages:
            yield message

log = Log()

def main():
    directory = make_temporary_directory()
    form = cgi.FieldStorage()
    inData = get_inData(form)
    user_workbook_name = log_file_name(inData)
    local_input_file_name = directory.name + '/inData.xlsx'
    write_file_to_temp(inData.file, local_input_file_name)
    local_output_file_name = directory.name + '/results.xlsx'
    wb = WB(log, directory.name)
    wb.load_workbook(local_input_file_name)
    wb.write(local_output_file_name)
    return_file = open_return_file(local_output_file_name)
    stdout.buffer.write(b'Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\r\n')
    stdout.buffer.write(bytes(
        'Content-Disposition: attachment; filename="{}"\r\n'.format(
            'results-' + user_workbook_name), encoding='utf-8'))
    stdout.buffer.write(b'\r\n')
    copyfileobj(return_file, stdout.buffer)

def open_return_file(local_output_file_name):
    try:
        return open(local_output_file_name, 'rb')
    except Exception as e:
        abort('The Pontogram service cannot read its own results file. '
            'The server says: {}'
            .format(e))

def make_temporary_directory():
    try:
        return TemporaryDirectory(dir='Temp')
    except Exception as e:
        abort('The Pontogram service cannot allocate a working area on the '
            'server computer. This is Not Your Fault. The server says: {}'
            .format(e))

def get_inData(form):
    if 'inData' not in form:
        abort('The data your browser sent does not include the required'
            ' "input data" field.')
    return form['inData']

def log_file_name(inData):
    file_name = inData.filename
    if not file_name:
        abort('No input file has been selected.')
    log.info('file_name = “{}”'.format(file_name))
    return file_name

def write_file_to_temp(file_input, local_file_name):
    if file_input:
        try:
            with open(local_file_name, 'wb') as sink:
                copyfileobj(file_input, sink)
        except Exception as e:
            abort('The Pontogram service cannot store a copy of your data.'
                'It says: {}'.format(e))
    else:
        abort('No file has been selected as the "input data"')

def abort(final_log_entry):
    stdout.buffer.write(b'Content-Type: text/html; charset=utf-8\r\n\r\n')
    stdout.buffer.write(bytes('''<!DOCTYPE html>
<html lang=en>
<meta charset=UTF-8>
<meta http-equiv="cache-control" content="max-age=0">
<meta http-equiv="cache-control" content="no-cache">
<meta http-equiv="expires" content="-1">
<meta http-equiv="expires" content="Tue, 01 Jan 1980 11:00:00 GMT">
<meta http-equiv="pragma" content="no-cache">
<meta name=description content="Error message from Pontogram servuce">
<link rel=stylesheet href=../pontogram.css>
<title>Pontogram Fatal Error</title>
''', encoding='utf-8'))
    log.error(final_log_entry)
    for severity, text in log.all_messages():
        escaped = html.escape(str(text))
        stdout.buffer.write(bytes('<p class={}>{}\n'.format(
            severity, escaped), encoding='utf-8'))
    for line in format_exc().split('\n'):
        escaped = html.escape(line)
        stdout.buffer.write(bytes('<p>{}\n'.format(escaped), encoding='utf-8'))
    stdout.buffer.flush()
    exit(0)

try:
    main()
except Exception as e:
    abort('Unexpected error. Server says: {}'.format(e))

