#! python3
# -*- mode: Python3; tab-width: 4; indent-tabs-mode: nil -*-

from collections import Counter, OrderedDict
from datetime import datetime
from numbers import Real
from pathlib import Path
import regex
from regex import fullmatch, sub
from subprocess import PIPE, run
from sys import stderr
from traceback import format_exc

import openpyxl

from xlsx import Xlsx

# Global log:
glog = None

class WB:

    def __init__(self, log=None, temp_dir='TEMP'):
        global glog
        glog = log
        self.dir = temp_dir  # str naming directory we can write temp files to
        self.grams = {1: None, 2: None}  # monograms, digrams
        self.log_time_of_day()
    
    def log_time_of_day(self):
        now = datetime.utcnow()
        glog.info(now)
    
    def load_workbook(self, file_name):
        """ 
        file_name -- str locating the source .xlsx file 
        """
        xlsx = Xlsx(glog, self.dir)
        xlsx.read_file(file_name)
        self.main = xlsx.main
        self.sheet_names = {name.strip(): name
            for name in self.main.name_to_Sheet}
        #print(sheet_names, file=stderr)
        for name in self.sheet_names.values():
            if name not in {
                    'trials', 'subjects', 'monograms', 'digrams', 'options'}:
                glog.warn(
                    'Ignoring “{}”, which is not a Pontogram worksheet'.format(
                        name))
        self.options = Options(self.get_sheet('options'))
        self.grams[1] = Monograms(self.get_sheet('monograms'))
        self.grams[2] = Digrams(self.get_sheet('digrams'))
        self.subjects = Subjects(self.get_sheet('subjects'))
        self.trials = Trials(self.get_sheet('trials'))
        self.grams[2].build_digram_matcher(self.grams[1])
        self.subjects.ignore_own_letters = self.options.ignore_own_letters
        self.subjects.incorporate_trials(self.trials)
        for glen in 1, 2:
            self.subjects.parse(self.grams, glen)
            self.subjects.count(self.grams, glen)
            if not self.options.skip_R:
                self.subjects.correlate_with_corpus(self.grams, glen, self.dir)
        #self.dump()
    
    def get_sheet(self, name):
        internal_name = self.sheet_names.get(name)
        if internal_name is None:
            return None
        return self.main.name_to_Sheet[internal_name]
    
    def write(self, file_name):
        xlsxwb = openpyxl.Workbook(write_only=True)
        self.write_log_sheet(xlsxwb)
        self.grams[1].write_to_xlsx(xlsxwb)
        self.grams[2].write_to_xlsx(xlsxwb)
        self.subjects.write_to_xlsx(xlsxwb)
        self.trials.write_to_xlsx(xlsxwb)
        self.subjects.write_subj_mono(xlsxwb, self.grams)
        self.subjects.write_subj_di(xlsxwb, self.grams)
        xlsxwb.save(file_name)
    
    def write_log_sheet(self, xlsxwb):
        ws = xlsxwb.create_sheet(title='log')
        ws.append(('severity', 'message'))
        for entry in glog.all_messages():
            ws.append(entry)
    
    def dump(self):
        print('WB:', file=stderr)
        self.options.dump()
        self.grams[1].dump()
        self.grams[2].dump()
        self.subjects.dump()


class Options:

    def __init__(self, xlsx_sheet):
        self.ignore_own_letters = False
        #self.return_matrix = False
        self.skip_R = False
        if not xlsx_sheet:
            return
        xlsx_sheet.load()
        needed_columns = {'option', 'selection'}
        col_r_to_name, name_to_col_r = process_worksheet_header(
            xlsx_sheet, needed_columns)
        if 'option' not in name_to_col_r:
            return
        for row_r, cells in data_rows(xlsx_sheet, col_r_to_name):
            option = cells.get('option')
            if option is None:
                glog.warn(
                    'In sheet “options”, ignoring row with blank “option”')
                continue
            try:
                option = option.lower()
            except AttributeError:
                pass
            selection = cells.get('selection')
            try:
                selection = selection.lower()
            except AttributeError:
                pass
            if option == 'ignore-own-letters':
                if selection in {'all', 'first', 'except-first'}:
                    self.ignore_own_letters = selection
                else:
                    self.ignore_own_letters = read_bool(selection)
                    if self.ignore_own_letters:
                        self.ignore_own_letters = 'all'
                continue
            if option == 'return-matrix':
                glog.warn('In sheet “options”, ignoring option “return-matrix”.'
                    ' That option has been retired. See sheets “subj_mono” and '
                    '“subj_di” for similar functionality.')
                #self.return_matrix = read_bool(selection)
                continue
            if option == 'skip-r':
                self.skip_R = read_bool(selection)
                continue
            glog.warn('In sheet “options”, ignoring unknown option “{}”'.format(
                option))
        xlsx_sheet.unload()  # maybe spare some memory
    
    def dump(self):
        print('Options:', file=stderr)
        print('  ignore_own_letters =', self.ignore_own_letters, file=stderr)
        print('  skip_R =', self.skip_R, file=stderr)
        print(file=stderr)

def read_bool(xlsx_val):
    if xlsx_val in {True, 1, 'TRUE', 'yes', None}:
        return True
    if xlsx_val in {False, 0, 'FALSE', 'no'}:
        return False
    glog.warn('Expected TRUE or FALSE, not {}; treating as FALSE'.format(
        xlsx.val))
    return False


class Monograms:

    def __init__(self, xlsx_sheet):
        self.grams = None
        if not xlsx_sheet:
            glog.error('Workbook has no “monograms” sheet.')
            return
        self.grams = Grams('monogram')
        self.grams.load(xlsx_sheet)
        self.index = self.grams.index
        self.source_rows = self.grams.source_rows
        self.gramAlts = None
        for g, rec in self.index.items():
            rec.base = Monograms.strip_to_uncased_base(g)
        self.build_matcher()
        
    @staticmethod
    def strip_to_uncased_base(g):
        return sub('\p{M}', '', g.lower())
    
    def build_matcher(self):
        grams = [regex.escape(rec.gram, special_only=True)
            for rec in self.index.values()]
        if not grams:
            glog.warn('In sheet “monograms”, no monograms are defined')
            self.regexS = '(?<special>[? ])|(?<other>.)'
        else:
            gramL = sorted(grams, key=(lambda x: len(x)), reverse=True)
            self.gramAlts = '|'.join(gramL)
            regexS = (
                '(?<gram>{})|(?<special>[? ])|(?<other>.)'.format(
                self.gramAlts))
        self.matcher = regex.compile(regexS, flags=regex.V1|regex.F)
    
    def find_grams(self, text):
        if self.matcher is None:
            return
        for m in self.matcher.finditer(text):
            if m['other']:
                yield m['other']
                continue
            if m['special'] == '\N{SPACE}':
                continue
            if m['gram']:
                yield self.index[m['gram']]
    
    def write_to_xlsx(self, xlsxwb):
        if self.grams is None:
            return
        self.grams.write_to_xlsx(xlsxwb)
    
    def dump(self):
        self.grams.dump()
        if self.matcher is not None:
            print(self.matcher, file=stderr)
        

class Digrams:

    def __init__(self, xlsx_sheet):
        self.grams = Grams('digram')
        self.index = self.grams.index
        self.source_rows = self.grams.source_rows
        self.matcher = None
        if not xlsx_sheet:
            glog.error('Workbook has no “digrams” sheet.')
            return
        self.grams.load(xlsx_sheet)
        
    def build_digram_matcher(self, monograms):
        if monograms.gramAlts is None:
            return
        regexS = ('(?<gram>(?:{gram})(?:{gram}))|(?<special>[? ])|(?<other>.)'
            .format(gram=monograms.gramAlts))
        self.matcher = regex.compile(regexS, flags=regex.V1|regex.F)
        for rec in self.index.values():
            digram = rec.gram
            m = fullmatch(
                    '(?<g1>{gram})(?<g2>{gram})'.format(
                    gram=monograms.gramAlts),
                    digram, flags=regex.V1|regex.F)
            if not m:
                glog.error('In sheet “digrams”, digram “{}” is not a'
                    ' combination of two monograms'.format(digram))
                continue
            rec.decomp = (
                monograms.index[m['g1']],
                monograms.index[m['g2']])
            rec.hetero = 1 if rec.decomp[0].base != rec.decomp[1].base else 0
            rec.adjac_seq = 0
            if fullmatch('[a-y]|[0-8]', rec.decomp[0].base):
                if ord(rec.decomp[1].base) == ord(rec.decomp[0].base) + 1:
                    rec.adjac_seq = 1
    
    def find_grams(self, text):
        if self.matcher is None:
            return
        for m in self.matcher.finditer(text, overlapped=True):
            if m['gram']:
                yield self.index[m['gram']]
                #yield True, m['gram']
    
    def write_to_xlsx(self, xlsxwb):
        if self.grams is None:
            return
        self.grams.write_to_xlsx(xlsxwb)
    
    def dump(self):
        self.grams.dump()
        if self.matcher is not None:
            print(self.matcher, file=stderr)


class Grams:

    def __init__(self, key_head):
        self.key_head = key_head
        self.freq_usable = False  # has 1 or more rows, all with a  Real freq
        self.covar_usable = False  # every row with a freq has a covar
        self.source_rows = []  # aligns with user source table: each row is
            # a gram or None
        self.index = OrderedDict()
        
    def load(self, xlsx_sheet):
        xlsx_sheet.load()
        col_r_to_name, name_to_col_r = process_worksheet_header(
            xlsx_sheet, {self.key_head, 'freq', 'covar'})
        for header in self.key_head, 'freq':
            if header not in name_to_col_r:
                glog.error(
                    'In sheet “{}”, required column “{}” is missing'
                    .format(xlsx_sheet.name, header))
            else:
                self.freq_usable = True
        if 'covar' in name_to_col_r:
            self.covar_usable = True
        last_row_r = 1  # the header row
        for row_r, cells in data_rows(xlsx_sheet, col_r_to_name):
            for r in range(last_row_r + 1, row_r):
                self.source_rows.append(None)
            gram_data = None
            gram = cells.get(self.key_head)
            if gram is not None:
                gram_data = GramData(gram)
                if not isinstance(gram, str):
                    glog.error('In sheet “{}”, gram “{}” is not textual'.format(
                        xlsx_sheet.name, gram))
                    gram = str(gram)
            if gram_data and gram in self.index:
                glog.error('In sheet “{}”, entries in column “{}” should be'
                    ' unique, but “{}” is repeated; only the last one will be'
                    ' used'.format(xlsx_sheet.name, key_head, gram))
                    
            freq = cells.get('freq')
            if not isinstance(freq, Real):
                if 'freq' in name_to_col_r:
                    glog.error('In sheet “{}”, gram “{}” doesn’t have a '
                        'number in the “freq” column'.format(
                        xlsx_sheet.name, key_head, gram))
                gram_data = None
            if gram_data and not isinstance(freq, Real):
                glog.error('Skipping all frequency calculations')
                gram_data = None
            if gram_data:
                gram_data.freq = freq
            if 'covar' in name_to_col_r:
                covar = cells.get('covar')
                if not isinstance(covar, Real):
                    glog.error('In sheet “{}”, the {} “{}” doesn’t have a '
                        'number in the “covar” column'.format(
                        xlsx_sheet.name, key_head, gram))
                    self.covar_usable = False
                    if gram:
                        glog.error('Skipping all use of covars')
                        gram_data = None
                elif gram_data:
                    gram_data.covar = covar
            self.source_rows.append(gram_data)
            if gram_data is not None:
                self.index[gram_data.gram] = gram_data
            last_row_r = row_r
        if not self.index:
            self.freq_usable = False
        if not self.freq_usable:
            self.covar_usable = False
        if self.freq_usable:
            total_freq = 0
            for gram, gram_data in self.index.items():
                total_freq += gram_data.freq
            if total_freq > 0:
                for gram, gram_data in self.index.items():
                    gram_data.corpProp = gram_data.freq / total_freq
        else:
            for gram, gram_data in self.index.items():
                gram_data.freq = None
        if not self.covar_usable:
            for gram, gram_data in self.index.items():
                gram_data.covar = None
        xlsx_sheet.unload()  # maybe spare some memory
    
    def yield_gram_freqs(self):
        for rec in self.index.values():
            yield rec.freq, rec.gram
            
    def yield_covar_counts(self):
        for rec in self.index.values():
            yield rec.covar, rec.gram
    
    def write_to_xlsx(self, xlsxwb):
        ws = xlsxwb.create_sheet(title=self.key_head + 's')
        header = [self.key_head]
        if self.key_head == 'monogram':
            header.append('base')
        else:
            header.extend(['decomp', 'transit', 'alpha'])
        if self.freq_usable:
            header.append('freq')
            header.append('prop')
        if self.covar_usable:
            header.append('covar')
        ws.append(header)
        for gram_data in self.source_rows:
            if gram_data is None:
                ws.append(())
                continue
            gram_data.write_to_xlsx(ws, header)
            
    def dump(self):
        print('{}:'.format(self.key_head), file=stderr)
        for gram_data in self.index.values():
            print('  ', end='', file=stderr)
            gram_data.dump()
            

class GramData:

    def __init__(self, gram):
        self.gram = gram
        self.base = None
        self.decomp = None  # digrams only
        self.hetero = 0  # digrams only
        self.adjac_seq = 0  # digrams only
        self.freq = None
        self.corpProp = None
        self.covar = None
    
    def __repr__(self):
        return '<Gram {}>'.format(self.gram)
    
    def write_to_xlsx(self, ws, header):
        row = []
        for h in header:
            row.append(
                self.gram if h in {'monogram', 'digram'} else
                self.base if h == 'base' else
                ' '.join(grec.base for grec in self.decomp) if h == 'decomp' 
                    else
                self.hetero if h == 'transit' else
                self.adjac_seq if h == 'alpha' else
                self.freq if h == 'freq' else
                self.corpProp if h == 'prop' else
                self.covar if h == 'covar' else '?')
        ws.append(row)
    
    def dump(self):
        print('“{}”'.format(self.gram), end='', file=stderr)
        if self.freq is not None:
            print('    freq = {}'.format(self.freq), end='', file=stderr)
        if self.covar is not None:
            print(', covar = {}'.format(self.covar), end='', file=stderr)
        if self.corpProp is not None:
            print(', corpProp = {}'.format(self.corpProp), file=stderr)
        if self.decomp is None:
            print('    base:', self.base, file=stderr)
        else:
            print('    decomp = {}'.format(self.decomp), end='', file=stderr)
            print(', hetero = {}'.format(self.hetero), end='', file=stderr)
            print(', adjac_seq = {}'.format(self.adjac_seq),
                file=stderr)


class Subjects:

    def __init__(self, xlsx_sheet):
        self.source_rows = []  # aligns with user source table: each row is
            # a Subject or None
        self.index = OrderedDict()
        if not xlsx_sheet:
            glog.error('Workbook has no “subjects” sheet.')
            return
        xlsx_sheet.load()
        col_r_to_name, name_to_col_r = process_worksheet_header(
            xlsx_sheet, {'subject', 'name'})
        last_row_r = 1  # header
        for row_r, cells in data_rows(xlsx_sheet, col_r_to_name):
            #TODO: have data_rows translate col_r to column headers
            #TODO: Move this mostly to Subject
            for r in range(last_row_r + 1, row_r):
                self.source_rows.append(None)
            subjectID = cells.get('subject')
            if subjectID is not None and not isinstance(subjectID, str):
                glog.error('In sheet “subjects”, “{}” is not textual'.format(
                    subjectID))
                subjectID = str(subjectID)
            if subjectID is not None:
                subject = Subject(subjectID)
                if subjectID in self.index:
                    glog.error('In sheet “subjects”, entries in column'
                        ' “subject” should be unique, but “{}” is repeated;'
                        ' only the last one will be used'.format(subjectID))
            name = cells.get('name')
            if name is not None and not isinstance(name, str):
                glog.error('In sheet “subjects”, subject “{}”, name “{}” is not'
                    ' textual'.format(name))
                name = str(name)
            if subject:
                subject.name = name
            self.source_rows.append(subject)
            if subject:
                self.index[subject.id] = subject
            last_row_r = row_r
        xlsx_sheet.unload()  # maybe spare some memory
    
    def incorporate_trials(self, trials):
        for trial in trials.index.values():
            subject_id = trial.subjectID
            if subject_id not in self.index:
                subject = Subject(subject_id)
                self.index[subject_id] = subject
                self.source_rows.append(subject)
            subject = self.index[subject_id]
            subject.trials.append(trial)
    
    def parse(self, all_grams, glen):
        grams = all_grams[glen]
        if grams.matcher is None:
            return
        for subject in self.index.values():
            subject.parse(all_grams, glen, self.ignore_own_letters)
    
    def count(self, all_grams, glen):
        for subject in self.index.values():
            subject.count(all_grams, glen)
    
    def correlate_with_corpus(self, all_grams, glen, working_dir):
        grams = all_grams[glen]
        if not grams.grams.freq_usable:
            return
        script_R_Path = self.write_R_script(grams, glen, working_dir)
        script_Rout_Path = self.run_R_script(script_R_Path)
        if script_Rout_Path:
            self.read_Rout(script_Rout_Path, glen)
    
    def write_R_script(self, grams, glen, working_dir):
        script_R_Path = Path(working_dir, 'script{}.R'.format(glen))
        #print(script_R_Path, file=stderr)
        with script_R_Path.open('w', encoding='utf-8') as sinkR:
            if grams.grams.covar_usable:
                print('library(ppcor)', file=sinkR)
            print('corpus_freqs <- c(', file=sinkR)
            write_numbers_to_R(grams.grams.yield_gram_freqs(), sinkR)
            print(')', file=sinkR)
            if grams.grams.covar_usable:
                print('covar <- c(', file=sinkR)
                write_numbers_to_R(grams.grams.yield_covar_counts(), sinkR)
                print(')', file=sinkR)
            for subject in self.index.values():
                subject.write_to_R_script(sinkR, grams, glen)
            print('quit(save="no", status=0)', file=sinkR)
        return script_R_Path
    
    def run_R_script(self, input_path):
        output_path = input_path.with_suffix('.Rout')
        cmd = 'R', 'CMD', 'BATCH', str(input_path), str(output_path)
        completion_status = run(cmd)
        if completion_status.returncode != 0:
            glog.error('R program failed to run; return code {}'.format(
                completion_status.returncode))
            return None
        return output_path
    
    def read_Rout(self, script_Rout_Path, glen):
        with script_Rout_Path.open(encoding='utf-8') as src:
            for line in (line.rstrip() for line in src.readlines()):
                #SubjID S100 corr 0.256348 p 0.06246775 
                m = fullmatch(r'SubjID (?<subject>.+?)'
                    r' corr (?<corr>.+?) p (?<p>.+?)', line)
                if m:
                    subject = self.index.get(m['subject'])
                    if subject is None:
                        glog.error('R returned a subject ID that we didn’t '
                            'give it: “{}”'.format(m['subject']))
                        continue
                    subject.stats[glen].add_corr_results(m['corr'], m['p'])
    
    def write_to_xlsx(self, xlsxwb):
        ws = xlsxwb.create_sheet(title='subjects')
        header = [
            'subject', 'name', 'nTrials', 
            'mono_name_set', 'mono_count', 'mono_freq', 'mono_avfreq',
                'mono_idio_count', 'mono_idio_prop',
                'mono_corr', 'mono_corr_p',
            'di_name_set', 'di_count', 'di_freq', 'di_avfreq',
                'di_idio_count', 'di_idio_prop',
                'di_corr', 'di_corr_p',
                'tran_count', 'tran_prop', 'alph_count', 'alph_prop']
        ws.append(header)
        for subject in self.source_rows:
            if subject is None:
                ws.append(())
            else:
                subject.write_row_to_xlsx(ws)

    def write_subj_mono(self, xlsxwb, grams):
        if grams[1].grams:
            ws = xlsxwb.create_sheet(title='subj_mono')
            header = 'subject', 'monogram', 'mono_count', 'mono_prop'
            ws.append(header)
        for subject in self.source_rows:
            if subject is None:
                continue
            subject.write_subj_gram(ws, grams, 1)

    def write_subj_di(self, xlsxwb, grams):
        if grams[2].grams:
            ws = xlsxwb.create_sheet(title='subj_di')
            header = 'subject', 'digram', 'di_count', 'di_prop'
            ws.append(header)
        for subject in self.source_rows:
            if subject is None:
                continue
            subject.write_subj_gram(ws, grams, 2)
                
    def dump(self):
        print('subjects:', file=stderr)
        for subject in self.index.values():
            print('  ', end='', file=stderr)
            subject.dump()
        print(file=stderr)


def write_numbers_to_R(gen, sinkR):
    n_written = 0
    for freq, g in gen:
        if n_written:
            print(',', file=sinkR)
        print('  {}  # {}'.format(freq, g), file=sinkR)
        n_written += 1


class Subject:

    def __init__(self, _id):
        self.id = _id
        self.name = None
        self.trials = []
        self.stats = {1: SubjectStats(), 2: SubjectStats()}
        
    def parse(self, all_grams, glen, ignore_own_letters):
        grams = all_grams[glen]
        name = self.name
        name_set = set()
        first_letter = set()
        if name:
            for info in grams.find_grams(name):
                if isinstance(info, GramData):
                    name_set.add(info)
                    if not first_letter:
                        first_letter = {info}
                else:
                    glog.warn('In sheet “subjects”, subject “{}”, name “{}”'
                        ' contains graph not found in sheet “monograms”: {}'
                        .format(subjectID, name, info))
        ignore_own_exclusions = (frozenset() if not ignore_own_letters else
            frozenset(
                first_letter
                    if ignore_own_letters == 'first' else
                name_set.difference(first_letter)
                    if ignore_own_letters == 'except-first' else
                name_set))
        self.stats[glen].name_set = frozenset(name_set)
        for trial in self.trials:
            if not trial.production:
                continue
            parse = []
            for info in grams.find_grams(trial.production):
                if isinstance(info, GramData): 
                    if info in ignore_own_exclusions:
                        continue
                    parse.append(info)
                else:
                    glog.warn('In sheet “trials”, trial “{}”, production'
                        ' “{}” contains graph not found in sheet'
                        ' “monograms”: {}'
                        .format(trial.id, trial.production, info))
            trial.stats[glen].parse = parse
    
    def count(self, all_grams, glen):
        grams = all_grams[glen]
        subject_stats = self.stats[glen]
        for trial in self.trials:
            trial_stats = trial.stats[glen]
            trial_stats.count(subject_stats.name_set)
            subject_stats.add(trial_stats)
            subject_stats.add_grams_by_type(trial_stats)
        subject_stats.average()
    
    def write_to_R_script(self, sinkR, grams, glen):
        print('# subject', self.id, file=sinkR)
        print('subj_freqs <- c(', file=sinkR)
        write_numbers_to_R(self.stats[glen].yield_gram_counts(grams), sinkR)
        print(') ', file=sinkR)
        if grams.grams.covar_usable:
            print('answer <- pcor.test(subj_freqs, corpus_freqs, covar,'
                ' method="kendall")', file=sinkR)
        else:
            print('answer <- cor.test(subj_freqs, corpus_freqs, '
                'alternative="greater", method="kendall", exact=FALSE)',
                file=sinkR)
            # -- exact tests don’t work when there are ties
        print('cat("SubjID {}",'
            ' "corr", answer$estimate,'
            ' "p", answer$p.value, "\\n")'.format(self.id),
            file=sinkR)
    
    def write_row_to_xlsx(self, ws):
        row = [self.id, self.name, len(self.trials)]
        self.stats[1].append_to_xlsx_row(row, 1)
        self.stats[2].append_to_xlsx_row(row, 2)
        ws.append(row)

    def write_subj_gram(self, ws, grams, glen):
        self.stats[glen].write_subj_gram(ws, grams, glen, self.id)
    
    def dump(self):
        print('“{}”'.format(self.id), end='', file=stderr)
        if self.name:
            print(', name = “{}”'.format(self.name), file=stderr)
        print('    mono stats:', file=stderr)
        self.stats[1].dump()
        print('    di stats:', file=stderr)
        self.stats[2].dump()
        print('    trials:', file=stderr)
        for trial in self.trials:
            print('      ', end='', file=stderr)
            trial.dump()

class SubjectStats:
    #TODO: refactor so that SubjectStats and TrialStats share more

    def __init__(self):
        self.name_set = None
        self.gram_count = 0
        self.corpFreq = 0
        self.avCorpFreq = None
        self.idio_count = 0
        self.idio_prop = None
        self.hetero = 0  # di only
        self.hetero_prop = None  # di only
        self.adjac_seq = 0  # di only
        self.adjseq_prop = None  # di only
        self.per_gram = Counter()
        self.corr = None
        self.corr_p = None
        
    def add(self, trial_stats):
        self.gram_count += trial_stats.gram_count
        self.corpFreq += trial_stats.corpFreq
        self.idio_count += trial_stats.idio_count
        self.hetero += trial_stats.hetero
        self.adjac_seq += trial_stats.adjac_seq
    
    def average(self):
        if self.gram_count > 0:
            self.avCorpFreq = self.corpFreq / self.gram_count
            self.idio_prop = self.idio_count / self.gram_count
            self.adjseq_prop = self.adjac_seq / self.gram_count
            self.hetero_prop = self.hetero / self.gram_count
    
    def add_grams_by_type(self, trial_stats):
        if not trial_stats.parse:
            return
        for grec in trial_stats.parse:
            self.per_gram[grec] += 1
    
    def yield_gram_counts(self, grams):
        for rec in grams.index.values():
            yield self.per_gram[rec], rec.gram
            
    def add_corr_results(self, corr, corr_p):
        self.corr = corr
        self.corr_p = corr_p
    
    def append_to_xlsx_row(self, row, glen):
        #self.dump()
        name_set = set()
        if self.name_set:
            for grec in self.name_set:
                name_set.add(grec.gram if glen == 1 
                    else ''.join(part.gram for part in grec.decomp))
        row.extend([' '.join(sorted(name_set)),
            self.gram_count,
            self.corpFreq, self.avCorpFreq,
            self.idio_count, self.idio_prop,
            self.corr, self.corr_p])
        if glen == 2:
            row.extend([
                self.hetero, self.hetero_prop,
                self.adjac_seq, self.adjseq_prop])

    def write_subj_gram(self, ws, grams, glen, subj_id):
        for grec in grams[glen].grams.source_rows:
            if not grec:
                continue
            if grec not in self.per_gram:
                continue
            count = self.per_gram[grec]
            prop = count / self.gram_count
            row = subj_id, grec.gram, count, prop
            ws.append(row)
    
    def dump(self):
        print('      name_set:', self.name_set, file=stderr)
        print('      gram_count:', self.gram_count, file=stderr)
        if self.gram_count:
            print('      corpFreq:', self.corpFreq,
                    ', avCorpFreq', self.avCorpFreq, file=stderr)
            print('      idio_count:', self.idio_count,
                    ', idio_prop', self.idio_prop, file=stderr)
            print('      hetero:', self.hetero,
                    ', hetero_prop', self.hetero_prop, file=stderr)
            print('      adjac_seq:', self.adjac_seq,
                    ', adjseq_prop', self.adjseq_prop, file=stderr)
            if self.corr is not None:
                print('      τb:', self.corr,
                    ', p:', self.corr_p, file=stderr)
    

class Trials:

    def __init__(self, xlsx_sheet):
        self.source_rows = []
        self.index = OrderedDict()
        if not xlsx_sheet:
            glog.error('Workbook has no “trials” sheet.')
            return
        xlsx_sheet.load()
        col_r_to_name, name_to_col_r = process_worksheet_header(
            xlsx_sheet, {'trial', 'subject', 'production'})
        #print(col_r_to_name, name_to_col_r, file=stderr)
        last_row_r = 1  # header
        for row_r, cells in data_rows(xlsx_sheet, col_r_to_name):
            for r in range(last_row_r + 1, row_r):
                self.source_rows.append(None)
            trial = Trial(cells, name_to_col_r)
            if trial.empty():
                 trial = None
            else:
                if trial.id is None:
                    trial.id = Trials.generate_trialID()
                if trial.subjectID is None:
                    trial.subjectID = '⁋anonymous'
            self.source_rows.append(trial)
            if trial is not None:
                self.index[trial.id] = trial
            last_row_r = row_r
        xlsx_sheet.unload()  # maybe spare some memory

    next_trialID = 0
    
    @staticmethod
    def generate_trialID():
        trialID = '⁋trial{:04}'.format(Trials.next_trialID)
        Trials.next_trialID += 1
        return trialID
    
    def write_to_xlsx(self, xlsxwb):
        ws = xlsxwb.create_sheet(title='trials')
        header = (
            'trial', 'subject', 'production', 
            'mono_parse', 'mono_count', 'mono_freq', 'mono_avfreq',
                'mono_idio_count', 'mono_idio_prop',
            'di_parse', 'di_count', 'di_freq', 'di_avfreq',
                'di_idio_count', 'di_idio_prop',
                'tran_count', 'tran_prop', 'alph_count', 'alph_prop')
        ws.append(header)
        for trial in self.source_rows:
            if trial is None:
                ws.append(())
                continue
            trial.write_row_to_xlsx(ws)
    
    def dump(self):
        print('trials:', file=stderr)
        for trial in self.index.values():
            print('  ', end='', file=stderr)
            trial.dump()
            

class Trial:

    def __init__(self, cells, name_to_col_r):
        self.id = cells.get('trial')
        self.subjectID = cells.get('subject')
        self.production = cells.get('production')
        self.stats = {1: TrialStats(1), 2: TrialStats(2)}
        if self.id is not None:
            if not isinstance(self.id, str):
                glog.error('In sheet “trials”, trial ID “{}” is not textual'.
                    format(self.id))
                self.id = str(self.id)
        if self.subjectID is not None:
            if not isinstance(self.subjectID, str):
                glog.error('In sheet “trials”, row “{}”, subject ID “{}” is'
                    ' not textual'.format(self.id, self.subjectID))
                self.subjectID = str(self.subjectID)
        if self.production is not None:
            if not isinstance(self.production, str):
                glog.error('In sheet “trials”, row “{}”, production “{}” is'
                    ' not textual'.format(self.id, self.production))
    
    def empty(self):
        return (self.id is None and
            self.subjectID is None and 
            self.production is None)
    
    def write_row_to_xlsx(self, ws):
        row = [self.id, self.subjectID, self.production]
        self.stats[1].add_to_xlsx_row(row, 1)
        self.stats[2].add_to_xlsx_row(row, 2)
        ws.append(row)
    
    def dump(self):
        print('“{}”, production = “{}”'.format(
            self.id, (self.production or '')), file=stderr)
        print('        mono stats:', file=stderr)
        self.stats[1].dump()
        print('        di stats:', file=stderr)
        self.stats[2].dump()


class TrialStats:

    def __init__(self, glen):
        self.glen = glen
        self.parse = None
        self.gram_count = 0
        self.corpFreq = 0
        self.avCorpFreq = None
        self.idio_count = 0
        self.idio_prop = None
        self.hetero = 0
        self.hetero_prop = None
        self.adjac_seq = 0
        self.adjseq_prop = None
    
    def count(self, name_set):
        if not self.parse:
            return
        for grec in self.parse:
            self.gram_count += 1
            self.corpFreq += grec.freq or 0
            if grec in name_set:
                self.idio_count += 1
            self.hetero += grec.hetero
            self.adjac_seq += grec.adjac_seq
        if self.gram_count > 0:
            self.avCorpFreq = (self.corpFreq / self.gram_count)
            self.idio_prop = (self.idio_count / self.gram_count)
            self.adjseq_prop = self.adjac_seq / self.gram_count
            self.hetero_prop = self.hetero / self.gram_count
    
    def add_to_xlsx_row(self, row, glen):
        row.append(
            ' '.join(grec.gram for grec in self.parse) if self.parse else None)
        row.extend([
            self.gram_count,
            self.corpFreq, self.avCorpFreq,
            self.idio_count, self.idio_prop])
        if glen == 2:
            row.extend([
                self.hetero, self.hetero_prop,
                self.adjac_seq, self.adjseq_prop])
    
    def dump(self):
        print('          parse: ', self.parse, file=stderr)
        print('          gram_count:', self.gram_count, file=stderr)
        if self.gram_count:
            print('          corpFreq:', self.corpFreq, 
                    ', avCorpFreq:', self.avCorpFreq, file=stderr)
            print('          idio_count:', self.idio_count, 
                    ', idio_prop:', self.idio_prop, file=stderr)
            if self.glen > 1:
                print('          hetero:', self.hetero,
                        ', hetero_prop', self.hetero_prop, file=stderr)
                print('          adjac_seq:', self.adjac_seq,
                        ', adjseq_prop', self.adjseq_prop, file=stderr)
        

def process_worksheet_header(xlsx_sheet, expected_cols):
    col_r_to_name = {}
    name_to_col_r = {}
    row_r, top_row_ws = xlsx_sheet.read_top_row()
    if not top_row_ws:
        glog.warn('Sheet {} lacks top row header; ignoring'.format(
            xlsx_sheet.name))
        return col_r_to_name, name_to_col_r
    duplicates = set()
    encountered = set()
    for col_r, head in top_row_ws.items():
        if isinstance(head, str):
            head = head.strip()
        head = head.strip()
        if head in expected_cols:
            if head in encountered:
                duplicates.add(head)
            else:
                encountered.add(head)
                col_r_to_name[col_r] = head
                name_to_col_r[head] = col_r
        else:
            glog.warn('In sheet “{}”, ignoring column headed “{}”'.format(
                xlsx_sheet.name, head))
    for head in duplicates:
        glog.warn('In sheet “{}”, header “{}” is repeated.'
            .format(xlsx_sheet.name, head))
    for head in expected_cols:
        if head not in encountered:
            glog.warn('In sheet “{}”, expected column “{}” is lacking'.format(
                xlsx_sheet.name, head))
    return col_r_to_name, name_to_col_r

def data_rows(xlsx_sheet, col_r_to_name):
    relabeled = {}
    for row_r, cells in xlsx_sheet.all_rows():
        if row_r == 1:  # top row is a header
            continue
        for h, val in cells.items():
            if h not in col_r_to_name:
                continue
            if isinstance(val, str):
                val = val.strip()
            if val in {'---', '—'}:
                val = None
            relabeled[col_r_to_name[h]] = val
        yield row_r, relabeled

def OLD_get_col_val(cells, name_to_col_r, header):
    if not cells:
        return None
    if not name_to_col_r:
        return None
    if not header:
        return None
    col_r = name_to_col_r.get(header)
    if not col_r:
        return None
    return cells.get(col_r)
