#! python3
# -*- mode: Python3; indent-tabs-mode: nil; tab-width: 4 -*-

"""
xlsx module
"""
'''TODO:
Go ahead and xfer data from the xml. But keep organization congruent with
file organization as much as possible. Every file is represented by a File structure, which has a location (maybe factored as unzip-dir + relative-dir + file-name).

Bootstrap:
1. Load Content_Type data from special path name [Content_Types].xml
  (A file that itself has neither a type nor enters into Relationships.)
2. Load Relationships XMLPart from special path name _rels/.rels.
   I.e. ask Part factory to instantiate a part, of type defined from Content_Type, telling it the unzip-dir, path-parent='', and  path-base=''.
   Because Content_Type will return application/vnd.openxmlformats-package.relationships+xml, Part will make a Relationships object.
That object knows that any file it points to will be at {unzip-dir}/{path-parent}/_rels/{path-base}.rels. It will open the file and parse it, knowing its expected root, and other elements and attributes, along with their namespaces. As it parses, it makes stubs for each of the Relationships given -- i.e. everything that can be determined before actually loading the file.

Part - unzip-dir + path-parent + path-base; Content_Type.
    If it governs relationships, includes pointer to the Relationships Part.
  Package - top-level package directory, pointing to top-level _rels/.rels
        Relationships.
  XMLPart - knows how to use ElementTree to read and parse file content.
        May be done lazily. Inherited by the following, as defined by types
        in Content_Type (all in frame "application/vnd.openxmlformats-…+xml"
    ContentType - singleton, package content types declarator
            Has logic to tell part type given its full name relative to 
                unzip-dir (i.e. starting with /).
            Contains factory to create other File subtypes, below, given full
                name.
    Relationships - XMLFile declared as type package.relationships
            Keeps track of one ….rels file: the package-relative path of
                its *parent* _rels file, and all the relationships, indexed
                both by Type and Id
            This part itself cannot govern or depend on relationships. 
    CoreProperties - type package.core-properties
    CustomProperties - type officedocument.custom-properties
    ExtendedProperties- type officedocument.extended-properties
    Main - type officedocument.spreadsheetml.sheet.main
    Styles - type officedocument.spreadsheetml.styles
    Worksheet - type officedocument.spreadsheetml.worksheet
    SharedStrings - type officedocument.spreadsheetml.sharedStrings

How file names work:
* Segments separated by /
* File names starting with / are absolute references, i.e. start from the
    unzip directory. 
* Names not starting with / are relative to the stratum from which they are
    invoked: i.e. the directory of the file they are named in.
    However, those named in a *.rels file, which are always in a _rels
    directory, are relative to their parent directory.
* Relative refs can start with .. segments, to go up.
* If the last segment ends in .* that last part (minus the dot) is an extension
    used by [Content_Types].xml to type files not listed there in full.
* [Content_Types].xml always uses absolute references. (If it can use a relative
    one, it would be relative to the top directory; so we could treat
    such as absolute, prepending a /
* *.rels files usu use relative references (but could be absolute?)
* File refs could in principle occur elsewheree, but probably not in xlsx.
* They are not quite the same as Posix paths:
** If a base name begins with a ., the whole base name is an extension.
** /abc is top of the unzip directory, but we need to start above it to 
  read files.

So one idea is to have file objects keep track of their reference directory: 
a list of segments P, plus a segment N giving its own name
If a main part with reference directory ['xl'] wants to read the relations
it governs, it looks for ['xl', '_rels', N.rels]
That Relations object will maintain reference directory ['xl'].
When it is consulted for file name Targets, relative targets are prepended with the reference directory.
To consult [Content_Types].xml, the segments are joined with / and a leading / is added.
To read a file, Path(temp_dir, *segments)
'''

from collections import defaultdict
import os
from numbers import Real
from pathlib import Path, PurePosixPath
from regex import fullmatch
from subprocess import PIPE, run
from sys import stderr
from tempfile import mkdtemp
from unicodedata import normalize
from xml.etree import ElementTree

glog = None

class Xlsx:

    def __init__(self, log, temp_dir=None):
        self.temp_dir = temp_dir
        global glog
        glog = log
    
    def read_file(self, src):
        try:
            self.unzip(src)
            self.read_Content_Types()
            self.read_package_rels()
            self.read_main_and_relations()
        except Exception:
            glog.error('Unable to read workbook file {}'.format(src))
            raise
    
    def unzip(self, src):
        if not self.temp_dir:
            self.temp_dir = mkdtemp(prefix='UnzippedWB', dir='.')
        #print('Unpacking into {}'.format(self.temp_dir), file=stderr)
        command = 'unzip', src, '-d', self.temp_dir
        result = run(command, stdout=PIPE, stderr=PIPE)
        #print(result.stdout.decode('UTF-8'), file=stderr)
        if result.stderr and b'cannot find or open' in result.stderr:
            glog.debug('“unzip” cannot find or read “{}”'.format(self.src))
            raise Exception('unzip_cannot_find')
        if result.stderr and (
            b'End-of-central-directory signature not found' in result.stderr):
            glog.debug('End-of-central-directory signature not found')
            glog.error('Selected input file is not an .xlsx file')
            raise Exception('unzip_signature')
        if result.returncode or result.stderr:
            glog.debug('unzip returncode {}, stderr={}'.format(
                result.returncode, result.stderr.decode('UTF-8')))
            raise Exception('unzip_error_code')
    
    def read_Content_Types(self):
        path = PartPath(self.temp_dir, (), '[Content_Types].xml', None)
        self.ContentTypes = ContentsTypeFile(path)
            
    def read_package_rels(self):
        path = PartPath(self.temp_dir, (), '', self.ContentTypes)
        self.rels = Relationships(path)
        Type = self.ContentTypes.get_Type(self.rels.actual_file_path)
        expected = 'application/vnd.openxmlformats-package.relationships+xml'
        if Type != expected:
            glog.warn('Package relationships file {} is not declared to be {}'
                .format(path, expected))
        self.rels.read()
        return True
    
    def read_main_and_relations(self):
        target = self.rels.get_Target_for_Type(
            'http://schemas.openxmlformats.org/officeDocument/2006/'
            'relationships/officeDocument')
        if not target:
            glog.error('Workbook lacks main office document component')
            raise Exception('no_officeDocument')
        Type = self.ContentTypes.get_Type(target)
        expected = ('application/vnd.openxmlformats-officedocument.'
                'spreadsheetml.sheet.main+xml')
        if Type != expected:
            glog.warn('Package main file {} is not declared to be {}'.format(
                target, expected))
        try:
            self.main = Main(target)
            self.main.read()
        except Exception:
            glog.error('Couldn’t read main officeDocument file')
            raise
    
    def dump(self):
        print(self.__dict__)


class PartPath:

    def __init__(self, unzip_path, ref_dir, file_path, content_types):
        """path_parent is a list of segments preceding the base_name.
            path_base may include an extension.
        """
        self.unzip_path = unzip_path
        self.content_types = content_types
        file_path_segs = file_path.split('/')
        #print('file_path_segs:', file_path_segs, file=stderr)
        if file_path.startswith('/'):
            # absolute path name; will yield initial '' in list
            self.ref_dir = tuple(file_path_segs[1:-1])
        else:
            # path name is relative to ref_dir:
            self.ref_dir = tuple(ref_dir) + tuple(file_path_segs[:-1])
        #print('self.ref_dir:', self.ref_dir, file=stderr)
        self.file_path = file_path_segs[-1]
        dot_pos = self.file_path.rfind('.')
        self.extension = None if dot_pos == -1 else self.file_path[dot_pos+1 :]
        #This is for look-up in [Content_Types].xml:
        self.absolute_path_str_lower = (
            '/' + '/'.join(self.ref_dir + (self.file_path,))).lower()
        self.fs_path = str(Path(
            unzip_path, *self.ref_dir, self.file_path))
    
    def copy(self, ref_dir=None, file_path=None):
        return PartPath(
            self.unzip_path,
            (ref_dir if ref_dir is not None else self.ref_dir),
            (file_path if file_path is not None else self.file_path),
            self.content_types)
    
    def __repr__(self):
        return '<PartPath {}>'.format(self.fs_path)


class ContentsTypeFile:

    ns = '{http://schemas.openxmlformats.org/package/2006/content-types}'

    def __init__(self, part_path):
        self.part_path = part_path
        tree = ElementTree.parse(self.part_path.fs_path)
        root = tree.getroot()
        self.PartName_to_Type = {}
        self.Extension_to_Type = {}
        for type_mapping in root:
            if type_mapping.tag == ContentsTypeFile.ns + 'Override':
                self.read_Override(self.PartName_to_Type, type_mapping)
            elif type_mapping.tag == ContentsTypeFile.ns + 'Default':
                self.read_Default(self.Extension_to_Type, type_mapping)
            else:
                glog.warn('Ignoring unknown element type in content-types file:'
                    ' “{}”'.format(type_mapping.tag))

    def read_Override(self, map_, Override):
        PartName = Override.get('PartName').lower()
        ContentType = Override.get('ContentType')
        if not ContentType:
            glog.warn('Ignoring content-types Override element that lacks '
                'ContentType')
            return
        map_[PartName] = ContentType

    def read_Default(self, map_, Default):
        Extension = Default.get('Extension')
        if not Extension:
            glog.warn('Ignoring content-types Default element that lacks '
                'Extension')
            return
        ContentType = Default.get('ContentType')
        if not ContentType:
            glog.warn('Ignoring content-types Default element that lacks '
                'ContentType')
            return
        map_[Extension.lower()] = ContentType
    
    def get_Type(self, part_path):
        file_path = part_path.absolute_path_str_lower
        Type = self.PartName_to_Type.get(file_path)
        if Type:
            return Type
        extension = part_path.extension
        if extension:
            Type = self.Extension_to_Type.get(extension)
        if not Type:
            glog.warn('content-types assigns no type to {}'.format(
                part_path.absolute_path_str_lower))
        return Type
    
    def dump(self):
        print(self.__dict__)


class Relationships:

    ns = '{http://schemas.openxmlformats.org/package/2006/relationships}'

    def __init__(self, part_path):
        self.part_path = part_path
        self.actual_file_path = part_path.copy(
            ref_dir=part_path.ref_dir + ('_rels',),
            file_path=part_path.file_path + '.rels')
    
    def read(self):
        tree = ElementTree.parse(self.actual_file_path.fs_path)
        root = tree.getroot()
        self.expected_root(root)
        self.list = []
        self.Id_to_Relationship = {}
        self.Type_to_Relationships = defaultdict(list)
        for Relationship_element in root:
            expected = Relationships.ns + 'Relationship'
            if Relationship_element.tag != expected:
                glog.warn('In {}, ignoring element that is {} instead of {}'
                    .format(self.actual_file_path, Relationship_element.tag,
                    expected))
                continue
            rel = Relationship(Relationship_element, self.part_path)
            self.list.append(rel)
            if rel.Id in self.Id_to_Relationship:
                glog.warn('Multiple relationships with Id {} in {}; ignoring '
                    'all but last.'.format(rel.Id, self.actual_file_path))
            self.Id_to_Relationship[rel.Id] = rel
            self.Type_to_Relationships[rel.Type].append(rel)
    
    def expected_root(self, root):
        expected = Relationships.ns + 'Relationships'
        if root.tag != expected:
            glog.error('{} should begin with root element {}, not {}'.format(
                self.part_path.absolute_path_str_lower, expected, root.tag))
    
    def get_Target_for_Type(self, relationship_Type):
        try:
            return self.Type_to_Relationships[relationship_Type][0].Target
        except Exception:
            return None
    
    def __repr__(self):
        return '<Relationships {}>'.format(self.part_path)


class Relationship:

    def __init__(self, element, part_path):
        self.Id = element.get('Id')
        if not self.Id:
            glog.debug('Relationship {} has no Id'.format(element))
        self.Type = element.get('Type')
        if not self.Type:
            glog.debug('Relationship {} has no Type'.format(self.Id))
        Target = element.get('Target')
        if not Target:
            glog.debug('Relationship {} has no Target'.format(self.Id))
        self.Target = part_path.copy(file_path=Target)
        self.TargetMode = element.get('TargetMode', 'Internal')
        
    def __repr__(self):
        return '<Relationship {}, {} → {} {}>'.format(
            self.Id, self.Type, self.TargetMode, self.Target)

class Main:
    """Main part of workbook."""
    
    ns = '{http://schemas.openxmlformats.org/spreadsheetml/2006/main}'
    #t_pattern = './/{}t'.format(ns)

    def __init__(self, part_path):
        self.part_path = part_path
        self.rels = Relationships(part_path)
        #print(self.rels.actual_file_path, file=stderr)
        self.shared_strings = []
        if os.path.exists(self.rels.actual_file_path.fs_path):
            try:
                self.rels.read()
            except Exception:
                glog.debug('Can’t read Relationships for main worksheet file')
                raise
            self.load_shared_strings()
            #print(self.shared_strings)
            ''' Other possible relations we're ignoring so far:
            Calculation Chain <calcChain> §12.3.1 (0, 1)
                sequence of cells involved in formula calculations
            Chartsheet <chartsheet> §12.3.2 (0+)
                a chart stored in its own sheet
            Connections <connections> §12.3.4 (0, 1)
                all connections to services
            Custom Property §12.3.5 (0+)
                anything
            Custom XML Mappings <MapInfo> §12.3.6 (0, 1)
            Dialogsheet <dialogsheet> §12.3.7 (0+)
            External Workbook Reference <externalLink> §12.3.9 (0+)
            Metadata <metadata> §12.3.10 (0, 1)
                OnLine Analytical Processing (OLAP) technology
            Pivot Table Cache Definition <pivotCacheDefinition> §12.3.12 (0+)
            Shared Workbook Revision Headers <headers> §12.3.16 (0, 1)
            Shared Workbook User Data <users> §12.3.18
            Styles <styleSheet> §12.3.20 (0, 1)
              relationship type: styles
            Volatile Dependencies <volTypes> §12.3.22 (1)
              real-time data
            '''
        else:
            self.rels = None
        
    def load_shared_strings(self):
        targ = self.rels.get_Target_for_Type(
            'http://schemas.openxmlformats.org/'
                'officeDocument/2006/relationships/sharedStrings')
        if not targ:
            return
        tree = ElementTree.parse(targ.fs_path)
        root = tree.getroot()
        #print(root, file=stderr)
        expected = Main.ns + 'sst'
        if root.tag != expected:
            glog.debug('sharedStrings file {} should have root {}, not {}'
                .format(targ, expected, root.tag))
        #t_pattern = './/{}t'.format(Main.ns)
        for si in root:
            if si.tag == Main.ns + 'si':
                #Ignore all formatting
                inner_text = ''.join(si.itertext())
                inner_text = normalize('NFD', inner_text)
                self.shared_strings.append(inner_text)
                continue
            print('{} in {}: don’t know about this element'.format(
                si.tag, targ.fs_path), file=stderr)
    
    def read(self):
        tree = ElementTree.parse(self.part_path.fs_path)
        root = tree.getroot()
        self.expected_root(root)
        """ attributes:  --ignore
          conformance="strict" (default= transitional)
        """
        self.sheets = []
        self.name_to_Sheet = {}
        for element in root:
            if element.tag in {
                Main.ns + 'bookViews',  # Part 1, 18.2.1
                    # collection of <workbookView> specifying window position,
                    #  filter options, and other configurations
                Main.ns + 'calcPr',  # Part 1, 18.2.2
                    # attributes name Calculation Properties of source appl
                Main.ns + 'customWorkbookViews',  # Part 1, 18.2.4
                Main.ns + 'definedNames',  # Part 1, 18.2.6
                    # collection of definedName
                Main.ns + 'externalReferences',  # Part 1, 18.2.9
                Main.ns + 'extLst',  # Part 1, 18.2.10
                Main.ns + 'fileRecoveryPr',  # Part 1, 18.2.11
                Main.ns + 'fileSharing',  # Part 1, 18.2.12
                Main.ns + 'fileVersion',
                    # attributes name source application and its versioning
                Main.ns + 'functionGroups', # Part 1, 18.2.15
                Main.ns + 'oleSize', # Part 1, 18.2.16
                    # embedded object server
                Main.ns + 'pivotCaches', # Part 1, 18.2.18
                Main.ns + 'smartTagPr', # Part 1, 18.2.21
                    #  attributes defining Smart Tag Properties
                Main.ns + 'smartTagTypes', # Part 1, 18.2.23
                Main.ns + 'webPublishing', # Part 1, 18.2.24
                Main.ns + 'webPublishObjects', # Part 1, 18.2.26
                Main.ns + 'workbookPr', # Part 1, 18.2.28
                Main.ns + 'workbookProtection' # Part 1, 18.2.29
            }:
                continue  # ignore these 
            if element.tag == Main.ns + 'sheets':
                self.parse_sheets(element)
                continue
            print('{} in workbook: don’t know this element'.format(element.tag),
                file=stderr)
            '''
            
            '''
    
    def parse_sheets(self, sheets):
        for element in sheets:
            if element.tag == Main.ns + 'sheet':
                #print(element, file=stderr)
                # <sheet name="monograms" sheetId="3" state="visible" r:id="rId4"/>
                #print(element.attrib, file=stderr)
                sheet_rel_ID = element.get(
                    '{http://schemas.openxmlformats.org/officeDocument/2006/'
                    'relationships}id')
                #print(sheet_rel_ID, file=stderr)
                relationship = self.rels.Id_to_Relationship[sheet_rel_ID]
                #print(relationship, file=stderr)
                expected_rel_type = (
                    'http://schemas.openxmlformats.org/'
                    'officeDocument/2006/relationships/worksheet')
                if relationship.Type != expected_rel_type:
                    glog('Expected relationship type {}, not {}'.format(
                        expected_rel_type, relationship.Type))
                sheet = Sheet(element, relationship.Target, self)
                self.name_to_Sheet[sheet.name] = sheet
                    # name should be unique
                self.sheets.append(sheet)
                continue
            print('{} in workbook>sheets: don’t know this element'.format(
                    element.tag),
                file=stderr)
    
    def expected_root(self, root):
        expected = Main.ns + 'workbook'
        if root.tag != expected:
            glog.debug('{} should begin with root element {}, not {}'.format(
                self.part_path.absolute_path_str_lower, expected, root.tag))


class Sheet:

    ns = '{http://schemas.openxmlformats.org/officeDocument/2006/relationships}'

    def __init__(self, element, part_path, workbook):
        self.part_path = part_path
        self.workbook = workbook
        self.name = element.get('name')
        if not self.name:
            glog.debug('Sheet at {} has no name attribute')
        self.name = self.name.strip()
        self.sheetId = element.get('sheetId')
        if not self.sheetId:
            glog.debug('Sheet at {} has no sheetId attribute')
        # also 'state', sheet visibility
        '''
        #TODO: read its relationships. These include:
        
        Comments <comments> §12.3.3  (0, 1) per worksheet
        Drawings <wsDr> §12.3.8 (0+)
        Pivot Table <pivotTableDefinition> §12.3.11 (0+)
        Query Table <queryTable> §12.3.14
          connects a table to an external data source (corpus data?)
        Single Cell Table Definitions <singleXmlCells> §12.3.19 (0, 1)
          how to map non-repeating elements from a custom XML file into cells
        Table Definition <table> §12.3.21 (0+)
          description of a single table and its autofilter information.
          The data for the table is stored in the corresponding Worksheet part.
          
        Worth exploring:
        (1) tables (but LibreOffice calc just has unnamed AutoFilter regions)
          *maybe* an alternative to equating tables with sheets.
        (2) maybe Query Table framework for importing corpus data from spell.
        '''
    
    def load(self):
        tree = ElementTree.parse(self.part_path.fs_path)
        root = tree.getroot()
        self.expected_root(root)
        column_label_to_def = {}
        for element in root:
            if element.tag in {
               Main.ns + 'cols',
                    # 18.3.1.17 column formatting
                Main.ns + 'dimension',
                    # 18.3.1.35 Worksheet Dimensions (optional)
                Main.ns + 'headerFooter',
                    # 18.3.1.46
                Main.ns + 'pageMargins',
                    # 18.3.1.62
                Main.ns + 'pageSetup',
                    # 18.3.1.63
                Main.ns + 'printOptions',
                    # 18.3.1.70
                Main.ns + 'sheetFormatPr',
                    # 18.3.1.81 Sheet Format Properties
                Main.ns + 'sheetPr',
                    # 18.3.1.82 Sheet Properties
                Main.ns + 'sheetViews',
                    # 18.3.1.88
            }:
                continue
            if element.tag == Main.ns + 'sheetData':
                self.sheetData_element = element
                continue
            glog.debug('In {}, don’t know how to process element {}'.format(
                self.part_path.fs_path, element.tag))
                
    def unload(self):
        self.sheetData_element = None
    
    def expected_root(self, root):
        expected = Main.ns + 'worksheet'
        if root.tag != expected:
            glog.debug('{} should begin with root element {}, not {}'.format(
                self.part_path.absolute_path_str_lower, expected, root.tag))
    
    def read_top_row(self):
        #print('sde=', self.sheetData_element, file=stderr)
        top_row = self.sheetData_element.find(Main.ns + 'row')
        if not top_row:
            return None, None
        #print(top_row, file=stderr)
        return self.read_cells(top_row)
    
    def all_rows(self):
        for row in self.sheetData_element:
            if row.tag != (
                '{http://schemas.openxmlformats.org/spreadsheetml/2006/main}'
                    'row'):
                glog.debug('Where <row> expected, got element {}'
                    .format(row.tag))
                continue
            yield self.read_cells(row)
        
    def read_cells(self, row):
        cells = {}
        for cell in row:  # should be only cell and extList
            if cell.tag != (
                '{http://schemas.openxmlformats.org/spreadsheetml/2006/'
                    'main}c'):
                glog.debug('Where <c> expected, got element {}'
                    .format(cell.tag))
            row_r, col_r = self.parse_location(cell)
            value = self.parse_value(cell)
            cells[col_r] = value
        return row_r, cells
    
    def parse_location(self, cell):
        location = cell.get('r')
        if not location:
            glog.debug('<c> lacks r= attribute')
            raise Exception
        loc_match = fullmatch('(?<col>[A-Z]+)(?<row>\d+)', location)
        if not loc_match:
            glog.debug('Cannot parse <c>’s r={} attribute'.format(location))
            raise Exception
        col_r = loc_match['col']
        row_r = int(loc_match['row'])
        return row_r, col_r
    
    def parse_value(self, cell):
        data_type = cell.get('t', 'n')
        if data_type == 's':  # shared string
            # <c r="A1" s="2" t="s"><v>1345</v>
            index = cell.find(Main.ns + 'v').text
            value = self.workbook.shared_strings[int(index)]
        elif data_type == 'inlineStr':
            value = normalize('NFD', ''.join(cell.itertext()))
            #value = normalize('NFD', ''.join(el.text
            #    for el in cell.findall(Main.t_pattern)).strip())
        elif data_type == 'b':  # boolean
            num = cell.find(Main.ns + 'v').text
            value = num != '0'
        elif data_type == 'n':  # number
            content = cell.find(Main.ns + 'v')
            if content is None:  # empty cell
                value = None
            else:
                num = content.text
                try:
                    value = int(num)
                except Exception:
                    value = float(num)
        else:
            glog.debug('Unknown data_type {}'.format(data_type))
            raise Exception('unknown_data_type')
        return value
 
    def __repr__(self):
        return '<Sheet {}>'.format(self.name)
